from typing import List, Tuple, Optional

from mavehgvs.variant import Variant
from mavehgvs.exceptions import MaveHgvsParseError

__all__ = ["parse_variant_strings"]


def parse_variant_strings(
    variants: List[str]
) -> Tuple[List[Optional[Variant]], List[Optional[str]]]:
    """Parse a list of MAVE-HGVS strings into Variant objects or error messages.

    Parameters
    ----------
    variants : List[str]
        List of MAVE-HGVS strings to parse.

    Returns
    -------
    Tuple[List[Optional[Variant]], List[Optional[str]]]
        Returns a pair of lists containing variants or error messages.

        Both lists have the same length as the input list.
        The first list contains Variant objects if the string was successfully parsed; else None.
        The second list contains None if the string was successfully parsed; else the error message.

    """
    valid = list()
    invalid = list()

    for s in variants:
        try:
            v = Variant(s)
        except MaveHgvsParseError as error:
            valid.append(None)
            invalid.append(str(error))
        else:
            valid.append(v)
            invalid.append(None)

    return valid, invalid



from mavehgvs.exceptions import MaveHgvsParseError
from mavehgvs.position import VariantPosition
from mavehgvs.variant import Variant
from mavehgvs.util import parse_variant_strings

__all__ = ["Variant", "VariantPosition", "MaveHgvsParseError", "parse_variant_strings"]



pos: str = r"[1-9][0-9]*"
"""str: Pattern matching a positive integer not starting with 0.

This pattern is used for sequence positions, as position 0 does not exist.
"""

pos_intron: str = rf"{pos}(?:[+-]{pos})?"
"""str: Pattern matching a position with optional intron component.

This pattern is used for sequence positions in an RNA or noncoding sequence.
"""

pos_intron_utr: str = rf"[*-]?{pos}(?:[+-]{pos})?"
"""str: Pattern matching a position with optional intron and UTR components.

This pattern is used for sequence positions in a coding sequence.
"""



import unittest

from mavehgvs.exceptions import MaveHgvsParseError
from mavehgvs.variant import Variant
from mavehgvs.position import VariantPosition


class TestCreateSingleVariantFromString(unittest.TestCase):
    def test_invalid_raises_error(self) -> None:
        valid_variant_strings = [
            "p.Glu27Trp",
            "c.122-6T>A",
            "g.44del",
            "c.78+5_78+10del",
            "c.77dup",
            "p.Pro12_Gly18dup",
            "p.Ala12_Pro13insGlyProCys",
            "r.22_23insauc",
            "c.43-6_595+12delinsCTT",
            "p.Ile71_Cys80delinsSer",
            "p.=",
            "c.=",
            "p.(=)",
        ]

        invalid_variant_strings = [
            "g.Glu27Trp",
            "p.27Glu>Trp",
            "p.122-6T>A",
            "G>A",
            "22G>A",
            "G.44del",
            "a.78+5_78+10del",
            "77dup",
            "n.Pro12_Gly18dup",
            "g.22_23insauc",
            "g.25_24del",
            "g.25_24ins",
            "r.43-6_595+12delinsctt",
            "x.=",
            "c.(=)",
        ]

        for s in valid_variant_strings:
            with self.subTest(s=s):
                Variant(s)  # should pass

        for s in invalid_variant_strings:
            with self.subTest(s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s)

    def test_sub(self) -> None:
        variant_strings = [
            "p.Glu27Trp",
            "p.Ter345Lys",
            "p.Cys22=",
            "g.48C>A",
            "c.122-6T>A",
            "c.*33G>C",
            "r.22g>u",
            "r.33+12a>c",
            "p.=",
            "p.(=)",
            "n.=",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

    def test_del(self) -> None:
        variant_strings = [
            "g.44del",
            "c.78+5_78+10del",
            "c.1_95del",
            "p.Gly18del",
            "p.Gln7_Asn19del",
            "r.34_36del",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

    def test_dup(self) -> None:
        variant_strings = [
            "g.22_24dup",
            "c.77dup",
            "c.101+1_101+7dup",
            "p.Pro12_Gly18dup",
            "p.Cys5dup",
            "r.12dup",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

    def test_ins(self) -> None:
        variant_strings = [
            "g.234_235insT",
            "c.84_85insCTG",
            "c.99+6_99+7insA",
            "p.His7_Gln8insSer",
            "p.Ala12_Pro13insGlyProCys",
            "r.22_23insauc",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

    def test_delins(self) -> None:
        variant_strings = [
            "g.22delinsAACG",
            "c.83_85delinsT",
            "c.43-6_595+12delinsCTT",
            "p.Ile71_Cys80delinsSer",
            "p.His44delinsValProGlyGlu",
            "r.92delinsgac",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

    def test_target_identical(self) -> None:
        variant_strings = [f"{prefix}.=" for prefix in "gmo" "cn" "r"]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertTrue(v.is_target_identical())

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))


class TestCreateMultiVariantFromString(unittest.TestCase):
    def test_creation(self):
        variant_strings = [
            "p.[Glu27Trp;Ter345Lys]",
            "p.[Gly18del;Glu27Trp;Ter345Lys]",
            "p.[Gln7_Asn19del;Glu27Trp;Ter345Lys]",
            "c.[1_35del;78+5_78+10del;122T>A]",
        ]

        invalid_variant_strings = [
            "p.[Glu27Trp;=;Ter345Lys]",
            "p.[(=);Gly18del;Glu27Trp;Ter345Lys]",
            "c.[12T>A;=;78+5_78+10del]",
        ]

        for s in variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertEqual(s, str(v))

        for s in invalid_variant_strings:
            with self.subTest(s=s):
                with self.assertRaises(MaveHgvsParseError):
                    v = Variant(s)

    def test_ordering(self):
        variant_string_tuples = [
            ("p.[Gly345Lys;Glu27Trp]", "p.[Glu27Trp;Gly345Lys]"),
            ("p.[Glu27Trp;Gly18del;Ter345Lys]", "p.[Gly18del;Glu27Trp;Ter345Lys]"),
            ("c.[122T>A;1_35del;78+5_78+10del]", "c.[1_35del;78+5_78+10del;122T>A]"),
        ]

        for s, _ in variant_string_tuples:
            with self.subTest(s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, relaxed_ordering=False)

        for s, s_ordered in variant_string_tuples:
            with self.subTest(s=s):
                # Should pass creation
                Variant(s, relaxed_ordering=True)

        for s, s_ordered in variant_string_tuples:
            with self.subTest(s=s):
                v = Variant(s, relaxed_ordering=True)
                self.assertEqual(s_ordered, str(v))

    def test_overlaps(self):
        invalid_variant_strings = [
            "p.[Glu27Trp;Glu27Trp]",
            "p.[Glu27Trp;Glu27Tyr]",
            "p.[Pro27Trp;Glu27Tyr]",
            "p.[Gly18del;Gly18Tyr]",
            "p.[Gln7_Asn19del;Glu13Trp]",
            "p.[Gln7_Asn19del;Glu13Trp;Ter345Lys]",
            "c.[1_95del;78+5_78+10del;122T>A]",
            "c.[1_95del;22T>A]",
            "n.[22G>A;22G>T]",
        ]

        for s in invalid_variant_strings:
            with self.subTest(s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s)


class TestCreateSingleVariantFromValues(unittest.TestCase):
    pass


class TestCreateMultiVariantFromValues(unittest.TestCase):
    pass


class TestTargetSequenceValidation(unittest.TestCase):
    def test_matching_dna_substitution(self):
        variant_tuples = [("ACGT", "c.1A>T"), ("ACGT", "c.3G>C")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                v = Variant(s, targetseq=target)
                self.assertEqual(s, str(v))

    def test_nonmatching_dna_substitution(self):
        variant_tuples = [("ACGT", "c.1C>T"), ("ACGT", "c.3T>C")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, targetseq=target)

    def test_valid_dna_del(self):
        variant_tuples = [("ACGT", "c.1_3del"), ("ACGT", "c.4del")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                v = Variant(s, targetseq=target)
                self.assertEqual(s, str(v))

    def test_invalid_dna_del(self):
        variant_tuples = [
            ("ACGT", "c.1_5del"),
            ("ACGT", "c.6_8del"),
            ("ACGT", "c.7del"),
        ]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, targetseq=target)

    def test_valid_dna_dup(self):
        variant_tuples = [("ACGT", "c.1_3dup"), ("ACGT", "c.4dup")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                v = Variant(s, targetseq=target)
                self.assertEqual(s, str(v))

    def test_invalid_dna_dup(self):
        variant_tuples = [
            ("ACGT", "c.1_5dup"),
            ("ACGT", "c.6_8dup"),
            ("ACGT", "c.7dup"),
        ]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, targetseq=target)

    def test_valid_dna_ins(self):
        variant_tuples = [("ACGT", "c.1_2insAAA"), ("ACGT", "c.3_4insT")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                v = Variant(s, targetseq=target)
                self.assertEqual(s, str(v))

    def test_invalid_dna_ins(self):
        variant_tuples = [
            ("ACGT", "c.4_5insA"),
            ("ACGT", "c.10_11insTCG"),
        ]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, targetseq=target)

    def test_valid_dna_delins(self):
        variant_tuples = [("ACGT", "c.1_2delinsA"), ("ACGT", "c.4delinsTAAGC")]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                v = Variant(s, targetseq=target)
                self.assertEqual(s, str(v))

    def test_invalid_dna_delins(self):
        variant_tuples = [
            ("ACGT", "c.4_5delinsA"),
            ("ACGT", "c.10_delinsTCG"),
        ]

        for target, s in variant_tuples:
            with self.subTest(target=target, s=s):
                with self.assertRaises(MaveHgvsParseError):
                    Variant(s, targetseq=target)


class TestMiscMethods(unittest.TestCase):
    def test_is_multi_variant(self):
        single_variant_strings = [
            "p.Glu27Trp",
            "c.122-6T>A",
            "g.44del",
            "c.78+5_78+10del",
            "c.77dup",
            "p.Pro12_Gly18dup",
            "p.Ala12_Pro13insGlyProCys",
            "r.22_23insauc",
            "c.43-6_595+12delinsCTT",
            "p.Ile71_Cys80delinsSer",
            "p.=",
        ]

        multi_variant_strings = []

        for s in single_variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertFalse(v.is_multi_variant())

        for s in multi_variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertTrue(v.is_multi_variant())

    def test_uses_extended_positions(self):
        non_extended_variant_strings = [
            "p.Glu27Trp",
            "g.44del",
            "c.77dup",
            "p.Pro12_Gly18dup",
            "p.Ala12_Pro13insGlyProCys",
            "r.22_23insauc",
            "r.22g>u",
            "p.Ile71_Cys80delinsSer",
            "p.=",
        ]

        extended_variant_strings = [
            "c.122-6T>A",
            "c.78+5_78+10del",
            "c.43-6_595+12delinsCTT",
            "c.*33G>C",
            "r.33+12a>c",
        ]

        for s in non_extended_variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertFalse(v.uses_extended_positions())

        for s in extended_variant_strings:
            with self.subTest(s=s):
                v = Variant(s)
                self.assertTrue(v.uses_extended_positions())


# TODO: multi-variant test cases
class TestMiscProperties(unittest.TestCase):
    def test_prefix(self):
        variant_tuples = [(prefix, f"{prefix}.=") for prefix in "gmo" "cn" "r"]

        for p, s in variant_tuples:
            with self.subTest(p=p, s=s):
                v = Variant(s)
                self.assertEqual(p, v.prefix)

    def test_variant_type(self):
        variant_tuples = [
            ("sub", "p.Glu27Trp"),
            ("sub", "c.122-6T>A"),
            ("del", "g.44del"),
            ("del", "c.78+5_78+10del"),
            ("dup", "c.77dup"),
            ("dup", "p.Pro12_Gly18dup"),
            ("ins", "p.Ala12_Pro13insGlyProCys"),
            ("ins", "r.22_23insauc"),
            ("delins", "c.43-6_595+12delinsCTT"),
            ("delins", "p.Ile71_Cys80delinsSer"),
        ]

        for t, s in variant_tuples:
            with self.subTest(t=t, s=s):
                v = Variant(s)
                self.assertEqual(t, v.variant_type)

    def test_position(self):
        variant_tuples = [
            (VariantPosition("Glu27"), "p.Glu27Trp"),
            (VariantPosition("122-6"), "c.122-6T>A"),
            (VariantPosition("44"), "g.44del"),
            ((VariantPosition("78+5"), VariantPosition("78+10")), "c.78+5_78+10del"),
            (VariantPosition("77"), "c.77dup"),
            ((VariantPosition("Pro12"), VariantPosition("Gly18")), "p.Pro12_Gly18dup"),
            (
                (VariantPosition("Ala12"), VariantPosition("Pro13")),
                "p.Ala12_Pro13insGlyProCys",
            ),
            ((VariantPosition("22"), VariantPosition("23")), "r.22_23insauc"),
            (
                (VariantPosition("43-6"), VariantPosition("595+12")),
                "c.43-6_595+12delinsCTT",
            ),
            (
                (VariantPosition("Ile71"), VariantPosition("Cys80")),
                "p.Ile71_Cys80delinsSer",
            ),
        ]

        for p, s in variant_tuples:
            with self.subTest(p=p, s=s):
                v = Variant(s)
                if isinstance(p, list):  # multi-variant
                    self.assertEqual(len(p), len(v.positions))
                    for q, vp in zip(p, v.positions):
                        if isinstance(q, tuple):
                            self.assertTupleEqual(q, vp)
                        else:
                            self.assertEqual(q, vp)
                if isinstance(p, tuple):
                    self.assertTupleEqual(p, v.positions)
                else:
                    self.assertEqual(p, v.positions)

    def test_sequence(self):
        variant_tuples = [
            (("Glu", "Trp"), "p.Glu27Trp"),
            (("T", "A"), "c.122-6T>A"),
            (None, "g.44del"),
            (None, "c.78+5_78+10del"),
            (None, "c.77dup"),
            (None, "p.Pro12_Gly18dup"),
            ("GlyProCys", "p.Ala12_Pro13insGlyProCys"),
            ("auc", "r.22_23insauc"),
            ("CTT", "c.43-6_595+12delinsCTT"),
            ("Ser", "p.Ile71_Cys80delinsSer"),
        ]

        for seq, s in variant_tuples:
            with self.subTest(seq=seq, s=s):
                v = Variant(s)
                self.assertEqual(seq, v.sequence)

    def test_target_id(self):
        variant_tuples = [
            (None, "p.Glu27Trp"),
            (None, "c.122-6T>A"),
            ("GeneX", "GeneX:p.Glu27Trp"),
            ("YFG1", "YFG1:c.122-6T>A"),
            ("ENST00000471181.7", "ENST00000471181.7:c.122-6T>A"),
            ("NM_007294.4", "NM_007294.4:c.122-6T>A"),
        ]

        for t, s in variant_tuples:
            with self.subTest(t=t, s=s):
                v = Variant(s)
                self.assertEqual(t, v.target_id)


if __name__ == "__main__":
    unittest.main()



from fqfa.constants import AA_CODES
from mavehgvs.patterns.util import combine_patterns, remove_named_groups
from mavehgvs.patterns.position import pos

amino_acid: str = rf"(?:{'|'.join(AA_CODES.values())})"
"""str: Pattern matching any amino acid or Ter.

This does not include ambiguous amino acids such as Glx and Xaa.
"""

aa_pos: str = rf"(?:{amino_acid}{pos})"
"""str: Pattern matching an amino acid code followed by a position.
"""

pro_sub: str = rf"(?P<pro_sub>(?:(?P<position>{aa_pos})(?P<new>{amino_acid}|=))|(?P<equal>=)|(?P<equal_sy>\(=\)))"
"""str: Pattern matching a protein substitution.
"""

pro_del: str = rf"(?P<pro_del>(?:(?P<start>{aa_pos})_(?P<end>{aa_pos})del)|(?:(?P<pos>{aa_pos})del))"
"""str: Pattern matching a protein deletion.
"""

pro_dup: str = rf"(?P<pro_dup>(?:(?P<start>{aa_pos})_(?P<end>{aa_pos})dup)|(?:(?P<pos>{aa_pos})dup))"
"""str: Pattern matching a protein duplication.
"""

pro_ins: str = rf"(?P<pro_ins>(?P<start>{aa_pos})_(?P<end>{aa_pos})ins(?P<seq>{amino_acid}+))"
"""str: Pattern matching a protein insertion.
"""

pro_delins: str = rf"(?P<pro_delins>(?:(?:(?P<start>{aa_pos})_(?P<end>{aa_pos}))|(?P<pos>{aa_pos}))delins(?P<seq>{amino_acid}+))"
"""str: Pattern matching a protein deletion-insertion.
"""

pro_variant: str = combine_patterns(
    [pro_sub, pro_del, pro_dup, pro_ins, pro_delins], None
)
"""str: Pattern matching any single protein variant event.
"""

pro_single_variant: str = rf"(?P<pro>p\.{pro_variant})"
"""str: Pattern matching any complete protein variant, including the prefix character.
"""

pro_multi_variant: str = rf"(?P<pro_multi>p\.\[{remove_named_groups(pro_variant)}(?:;{remove_named_groups(pro_variant)}){{1,}}\])"
"""str: Pattern matching any complete protein multi-variant, including the prefix character.

Named capture groups have been removed from the variant patterns because of non-uniqueness.
Another applications of single-variant regular expressions is needed to recover the named groups from each individual
variant in the multi-variant.
"""



import setuptools
import sys

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = ["fqfa>=1.2.0"]
# fqfa requires backported dataclasses in Python 3.6
if sys.version_info.major == 3 and sys.version_info.minor == 6:
    requirements.append("dataclasses")

setuptools.setup(
    name="mavehgvs",
    version="0.1.0",
    author="Daniel Esposito and Alan F Rubin",
    author_email="alan.rubin@wehi.edu.au",
    description=(
        "Regular expression-based validation of HGVS variant strings for clinical genetics and genomics applications."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VariantEffect/mavehgvs",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=requirements,
    test_suite="tests",
)



import unittest
import re
from mavehgvs.patterns.dna import (
    dna_sub_c,
    dna_sub_n,
    dna_sub_gmo,
    dna_del_c,
    dna_del_n,
    dna_del_gmo,
    dna_dup_c,
    dna_dup_n,
    dna_dup_gmo,
    dna_ins_c,
    dna_ins_n,
    dna_ins_gmo,
    dna_delins_c,
    dna_delins_n,
    dna_delins_gmo,
    dna_variant_c,
    dna_variant_n,
    dna_variant_gmo,
    dna_single_variant,
    dna_multi_variant,
)


class TestDnaSubC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_sub_c, flags=re.ASCII)

        cls.valid_strings = ["48C>A", "=", "122-6T>A", "*24G>C", "19+22A>G", "-27+3T>C"]

        cls.invalid_strings = ["22g>u", "48C>W", "22=", "122=/T>A"]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaSubN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_sub_n, flags=re.ASCII)

        cls.valid_strings = ["48C>A", "=", "122-6T>A", "19+22A>G"]

        cls.invalid_strings = [
            "22g>u",
            "48C>W",
            "22=",
            "122=/T>A",
            "*24G>C",
            "-27+3T>C",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaSubGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_sub_gmo, flags=re.ASCII)

        cls.valid_strings = ["48C>A", "="]

        cls.invalid_strings = ["122-6T>A", "22g>u", "48C>W", "22=", "122=/T>A", "0C>T"]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_del_c, flags=re.ASCII)

        cls.valid_strings = [
            "44del",
            "1_95del",
            "78+5_78+10del",
            "-25+1_-25+3del",
            "*17del",
        ]

        cls.invalid_strings = [
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_del_n, flags=re.ASCII)

        cls.valid_strings = ["44del", "1_95del", "78+5_78+10del"]

        cls.invalid_strings = [
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
            "-25+1_-25+3del",
            "*17del",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_del_gmo, flags=re.ASCII)

        cls.valid_strings = ["44del", "1_95del"]

        cls.invalid_strings = [
            "78+5_78+10del",
            "-25+1_-25+3del",
            "*17del",
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDupC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_dup_c, flags=re.ASCII)

        cls.valid_strings = [
            "22_24dup",
            "77dup",
            "101+1_101+7dup",
            "-25+1_-25+3dup",
            "*17dup",
        ]

        cls.invalid_strings = [
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDupN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_dup_n, flags=re.ASCII)

        cls.valid_strings = ["22_24dup", "77dup", "101+1_101+7dup"]

        cls.invalid_strings = [
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "-25+1_-25+3dup",
            "*17dup",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDupGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_dup_gmo, flags=re.ASCII)

        cls.valid_strings = ["22_24dup", "77dup"]

        cls.invalid_strings = [
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "101+1_101+7dup",
            "-25+1_-25+3dup",
            "*17dup",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaInsC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_ins_c, flags=re.ASCII)

        cls.valid_strings = [
            "234_235insT",
            "84_85insCTG",
            "*84_*85insCTG",
            "99+6_99+7insA",
            "124+100_124-100insTTG",
            "124+101_124-100insTTG",
        ]

        cls.invalid_strings = ["84_85ins100_125", "234_235ins(10)", "234_235ins(?)"]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaInsN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_ins_n, flags=re.ASCII)

        cls.valid_strings = [
            "234_235insT",
            "84_85insCTG",
            "99+6_99+7insA",
            "124+100_124-100insTTG",
            "124+101_124-100insTTG",
        ]

        cls.invalid_strings = [
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
            "*84_*85insCTG",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaInsGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_ins_gmo, flags=re.ASCII)

        cls.valid_strings = ["234_235insT", "84_85insCTG"]

        cls.invalid_strings = [
            "99+6_99+7insA",
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelinsC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_delins_c, flags=re.ASCII)

        cls.valid_strings = [
            "22delinsAACG",
            "83_85delinsT",
            "43-6_595+12delinsCTT",
            "*788delinsA",
        ]

        cls.invalid_strings = ["84_85delinsAAN", "234delinsW"]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelinsN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_delins_n, flags=re.ASCII)

        cls.valid_strings = ["22delinsAACG", "83_85delinsT", "43-6_595+12delinsCTT"]

        cls.invalid_strings = ["84_85delinsAAN", "234delinsW" "*788delinsA"]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaDelinsGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_delins_gmo, flags=re.ASCII)

        cls.valid_strings = ["22delinsAACG", "83_85delinsT"]

        cls.invalid_strings = [
            "43-6_595+12delinsCTT",
            "*788delinsA",
            "84_85delinsAAN",
            "234delinsW",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaVariantC(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_variant_c, flags=re.ASCII)

        cls.valid_strings = [
            "48C>A",
            "=",
            "122-6T>A",
            "*24G>C",
            "19+22A>G",
            "-27+3T>C",
            "44del",
            "1_95del",
            "78+5_78+10del",
            "-25+1_-25+3del",
            "*17del",
            "22_24dup",
            "77dup",
            "101+1_101+7dup",
            "-25+1_-25+3dup",
            "*17dup",
            "234_235insT",
            "84_85insCTG",
            "99+6_99+7insA",
            "22delinsAACG",
            "83_85delinsT",
            "43-6_595+12delinsCTT",
            "*788delinsA",
        ]

        cls.invalid_strings = [
            "22g>u",
            "48C>W",
            "22=",
            "122=/T>A",
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
            "84_85delinsAAN",
            "234delinsW",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaVariantN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_variant_n, flags=re.ASCII)

        cls.valid_strings = [
            "48C>A",
            "=",
            "122-6T>A",
            "19+22A>G",
            "44del",
            "1_95del",
            "78+5_78+10del",
            "22_24dup",
            "77dup",
            "101+1_101+7dup",
            "234_235insT",
            "84_85insCTG",
            "99+6_99+7insA",
            "22delinsAACG",
            "83_85delinsT",
            "43-6_595+12delinsCTT",
        ]

        cls.invalid_strings = [
            "22g>u",
            "48C>W",
            "22=",
            "122=/T>A",
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
            "84_85delinsAAN",
            "234delinsW",
            "*24G>C",
            "-27+3T>C",
            "-25+1_-25+3del",
            "*17del",
            "-25+1_-25+3dup",
            "*17dup",
            "*788delinsA",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaVariantGmo(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_variant_gmo, flags=re.ASCII)

        cls.valid_strings = [
            "48C>A",
            "=",
            "44del",
            "1_95del",
            "22_24dup",
            "77dup",
            "234_235insT",
            "84_85insCTG",
            "22delinsAACG",
            "83_85delinsT",
        ]

        cls.invalid_strings = [
            "43-6_595+12delinsCTT",
            "*788delinsA",
            "99+6_99+7insA",
            "101+1_101+7dup",
            "-25+1_-25+3dup",
            "*17dup",
            "78+5_78+10del",
            "-25+1_-25+3del",
            "*17del",
            "*24G>C",
            "19+22A>G",
            "122-6T>A",
            "-27+3T>C",
            "22g>u",
            "48C>W",
            "22=",
            "122=/T>A",
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
            "84_85delinsAAN",
            "234delinsW",
        ]

    def test_valid_strings(self):
        for s in self.valid_strings:
            with self.subTest(s=s):
                self.assertIsNotNone(
                    self.pattern.fullmatch(s), msg=f'failed to match "{s}"'
                )

    def test_invalid_strings(self):
        for s in self.invalid_strings:
            with self.subTest(s=s):
                self.assertIsNone(
                    self.pattern.fullmatch(s), msg=f'incorrectly matched "{s}"'
                )


class TestDnaSingleVariant(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pattern = re.compile(dna_single_variant, flags=re.ASCII)

        cls.valid_strings = [
            "48C>A",
            "=",
            "44del",
            "1_95del",
            "22_24dup",
            "77dup",
            "234_235insT",
            "84_85insCTG",
            "22delinsAACG",
            "83_85delinsT",
        ]

        cls.valid_strings_c_only = [
            "*788delinsA",
            "-25+1_-25+3dup",
            "*17dup",
            "-25+1_-25+3del",
            "*17del",
            "*24G>C",
            "-27+3T>C",
        ]

        cls.valid_strings_cn_only = [
            "43-6_595+12delinsCTT",
            "99+6_99+7insA",
            "101+1_101+7dup",
            "78+5_78+10del",
            "19+22A>G",
            "122-6T>A",
        ]

        cls.invalid_strings = [
            "22g>u",
            "48C>W",
            "22=",
            "122=/T>A",
            "(78+1_79-1)_(124+1_125-1)del",
            "(?_85)_(124_?)del",
            "122=/del",
            "(78+1_79-1)_(124+1_125-1)dup",
            "(?_85)_(124_?)dup",
            "122_125=//dup",
            "84_85ins100_125",
            "234_235ins(10)",
            "234_235ins(?)",
            "84_85delinsAAN",
            "234delinsW",
        ]

    def test_valid_strings(self):
        for p in "cngmo":
            for s in self.valid_strings:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNotNone(
                        self.pattern.fullmatch(v), msg=f'failed to match "{v}"'
                    )
        for p in "cn":
            for s in self.valid_strings_cn_only:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNotNone(
                        self.pattern.fullmatch(v), msg=f'failed to match "{v}"'
                    )
        for p in "c":
            for s in self.valid_strings_c_only:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNotNone(
                        self.pattern.fullmatch(v), msg=f'failed to match "{v}"'
                    )

    def test_invalid_strings(self):
        for p in "cngmo":
            for s in self.invalid_strings:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNone(
                        self.pattern.fullmatch(v), msg=f'incorrectly matched "{v}"'
                    )
        for p in "gmo":
            for s in self.valid_strings_cn_only:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNone(
                        self.pattern.fullmatch(v), msg=f'incorrectly matched "{v}"'
                    )
        for p in "n":
            for s in self.valid_strings_c_only:
                with self.subTest(s=s, p=p):
                    v = f"{p}.{s}"
                    self.assertIsNone(
                        self.pattern.fullmatch(v), msg=f'incorrectly matched "{v}"'
                    )


class TestDnaMultiVariant(unittest.TestCase):
    @unittest.expectedFailure
    def test_something(self):
        self.assertEqual(True, False)


if __name__ == "__main__":
    unittest.main()
